#ident "$Header$"
/*
Copyright (c) Members of the EGEE Collaboration. 2004-2010.
See http://www.eu-egee.org/partners for details on the copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/


/*


  !! Automatically generated file
  !! Do not edit, your changes will be discarded upon build
  !! Change the corresponding template file ./src/store.c.T


 * XXX: still lots of hardcoded stuff
 *	there's mapping db.column <-> event struct field
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>

#include "glite/jobid/strmd5.h"
#include "glite/lbu/trio.h"
#include "glite/lbu/log.h"

#include "glite/lb/events_parse.h"
#include "glite/lb/context-int.h"
#include "glite/lb/log_proto.h"	/* for EDG_WLL_LOG_USER_DEFAULT */
#include "glite/lb/intjobstat_supp.h"

#include "store.h"
#include "get_events.h"
#include "lb_authz.h"
#include "authz_policy.h"
#include "jobstat.h"
#include "db_calls.h"
#include "db_supp.h"
#include "index.h"

static int store_user(edg_wll_Context,const char *,const char *); 
static int store_job(edg_wll_Context,glite_jobid_const_t,const char *, int, int, int, int);
static int store_flesh(edg_wll_Context,edg_wll_Event *,const char *ulm, char *,int);
/* FIXME: 'check_dup' defined but not used */
static int check_dup(edg_wll_Context,edg_wll_Event *);
static int check_auth(edg_wll_Context,edg_wll_Event *e); 
static void lowercase_usertag(edg_wll_Event *ev);

void edg_wll_StoreAnonymous(edg_wll_Context ctx,int anon) {
	ctx->allowAnonymous = anon;
}


/* !!! to be called from OPEN TRANSACTION only !!! 
 */
int edg_wll_StoreEvent(edg_wll_Context ctx,edg_wll_Event *e,const char *ulm,int *seq)
{
	char			*userid, *jobid, *stmt, *ssrc, *now_s, *stamp, *dummy, *max;
	glite_lbu_Statement	sh = NULL;
	int			next = 0xDEAD, nr;


	userid = ssrc = jobid = stmt = now_s = stamp = dummy = max = NULL;

	lowercase_usertag(e);
	jobid = edg_wlc_JobIdGetUnique(e->any.jobId);
	glite_lbu_TimeToStr(e->any.timestamp.tv_sec, &stamp);
	ssrc = edg_wll_SourceToString(e->any.source);

	if ( ctx->event_load )
		glite_lbu_TimeToStr(e->any.arrived.tv_sec, &now_s);
	else
		glite_lbu_TimeToStr(time(NULL), &now_s);

	edg_wll_ResetError(ctx);
	switch (check_auth(ctx,e)) {
		case 0: break;
		case ENOENT: 
			/* job not registered */
			// should not happen, store_job_server_proxy() miscoded or going thu load?
			goto clean;
			break;
		case EPERM:
			if (!ctx->noAuth) goto clean;
			edg_wll_ResetError(ctx);
			break;
		default: goto clean;
	}

	trio_asprintf(&stmt,"select userid from jobs where jobid='%|Ss'", jobid);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);
	
	if (edg_wll_ExecSQL(ctx,stmt,&sh) < 0 || edg_wll_FetchRow(ctx,sh,1,NULL,&userid) < 0) goto clean;
	if (sh) glite_lbu_FreeStmt(&sh);
	free(stmt); stmt = NULL;


/* check duplicity */
	trio_asprintf(&stmt,
		"select arrived from events where jobid='%|Ss' and code='%d'"
		" and prog='%|Ss' and host='%|Ss' and time_stamp=%s and usec='%d'"
		" and level='%d' and userid='%|Ss' and seqcode='%|Ss'",
		jobid, (int) e->any.type,
		ssrc,e->any.host,
		stamp,e->any.timestamp.tv_usec,
		e->any.level,userid, e->any.seqcode);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,&sh) < 0) goto clean;
	nr = edg_wll_FetchRow(ctx,sh,1,NULL,&dummy);
	if (sh) glite_lbu_FreeStmt(&sh);
	free(stmt); stmt = NULL;
	free(dummy);

	if (nr < 0) goto clean;
	if (nr > 0) {
		/* possible duplicity (99%) */
		// XXX: check event flesh to be 100% sure
		edg_wll_SetError(ctx,EEXIST,"duplicate event");
		goto clean;
	}
	/* else (nr == 0) -> unique event, continue */
	

/* obtain number of stored events */
	trio_asprintf(&stmt,
		"select nevents from jobs "
		"where jobid = '%|Ss'",jobid);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,&sh) < 0 ||
	    edg_wll_FetchRow(ctx,sh,1,NULL,&max) < 0) goto clean;
	glite_lbu_FreeStmt(&sh); 
	free(stmt);
	
	next = (max && *max) ? atoi(max) : 0;
	free(max);


/* store event */		
	trio_asprintf(&stmt,
		"insert into events(jobid,event,code,prog,host,time_stamp,usec,arrived,level,userid,seqcode) "
		"values ('%|Ss',%d,%d,'%|Ss','%|Ss',%s,%d,%s,%d,'%|Ss','%|Ss')",
		jobid,next,
		(int) e->any.type,
		ssrc,e->any.host,
		stamp,e->any.timestamp.tv_usec,
		now_s, e->any.level,userid, e->any.seqcode);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) goto clean;
	free(stmt); stmt = NULL;

/* increase number of stored events */
	trio_asprintf(&stmt,
		"update jobs set nevents='%d'"
		"where jobid = '%|Ss'", next+1, jobid);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) goto clean;
	free(stmt); stmt = NULL;
	

/* store event record */
	if (store_flesh(ctx,e,ulm,jobid,next)) goto clean;

clean:
	free(now_s);
	free(userid);
	free(jobid);
	free(stmt);
	free(ssrc);
	if (sh) glite_lbu_FreeStmt(&sh);
	if (!edg_wll_Error(ctx,NULL,NULL) && seq) *seq = next;
	free(stamp);
	return edg_wll_Error(ctx,NULL,NULL);
}

static int store_user(edg_wll_Context ctx,const char *userid,const char *subj)
{
	char	*stmt;

	trio_asprintf(&stmt,"insert into users(userid,cert_subj) "
		"values ('%|Ss','%|Ss')",userid,subj);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) {
		if (edg_wll_Error(ctx,NULL,NULL) == EEXIST)
			edg_wll_ResetError(ctx);
	}

	free(stmt);
	return edg_wll_Error(ctx,NULL,NULL);
}

static int store_job(edg_wll_Context ctx,glite_jobid_const_t job,const char *userid, int proxy, int server,int grey, int update)
{
	char *jobstr = edg_wlc_JobIdUnparse(job);
	char *jobid = edg_wlc_JobIdGetUnique(job);
	char *stmt;
	char *srvName;
	unsigned int	srvPort;

/* debug Duplicate key on index: Duplicate entry '(nil)' for key 1
 */
	if (jobid == NULL || jobstr == NULL) 
		return edg_wll_SetError(ctx,EINVAL,"store_job()");

	edg_wll_ResetError(ctx);

	if (ctx->isProxy) {
		proxy = 1;

		/* If host&port in jobId match bkserver hostname and port and bkserver
		 * runs server service, mark the proxy job to belong to bkserver too	
		 */
		edg_wlc_JobIdGetServerParts(job, &srvName, &srvPort);
		if ( (ctx->serverRunning) && (ctx->srvPort == srvPort) && 
				!strcmp(ctx->srvName, srvName)) {
			server=1;
		}
	}
	else {
		server = 1;
	}

	if (update) {
		trio_asprintf(&stmt,"update jobs set userid='%|Ss', proxy='%|Sd', server='%|Sd', grey='%|Sd' where jobid='%|Ss'", 
			userid,proxy,server,grey,jobid);
	}
	else {
		trio_asprintf(&stmt,"insert into jobs(jobid,dg_jobid,userid,proxy,server,grey) "
			"values ('%|Ss','%|Ss','%|Ss', '%|Sd', '%|Sd', '%|Sd')",jobid,jobstr,userid,proxy,server,grey);
	}
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) {
		if (edg_wll_Error(ctx,NULL,NULL) == EEXIST && !update) 
			edg_wll_ResetError(ctx);
		else
			goto err;
	}
	free(stmt); stmt = NULL;

err:
	free(stmt);
	free(jobstr);
	free(jobid);
	return edg_wll_Error(ctx,NULL,NULL);
}


/* test whether job shares LB proxy and server DB or not */
int is_job_local(edg_wll_Context ctx, glite_jobid_const_t jobId) 
{
	char		*srvName = NULL;
	unsigned int	srvPort; 
	int		ret;	

	if (!ctx->serverRunning) return 0;	/* XXX */

	edg_wlc_JobIdGetServerParts(jobId, &srvName, &srvPort);
	ret = ((ctx->srvPort == srvPort) && ctx->srvName && !strcmp(srvName,ctx->srvName));
	free(srvName);

	return(ret);
}

int store_job_server_proxy(edg_wll_Context ctx, edg_wll_Event *event, int *register_to_JP)
{
	char 		*unique = edg_wlc_JobIdGetUnique(event->any.jobId);
	char            *q = NULL, *userid = NULL, *subj = NULL, *owner = NULL;
	glite_lbu_Statement    stmt = NULL, stmt_zomb = NULL;
	int             nar, grey = 0, async = 0;
	char		*can_peername = NULL;
	int             local_job = is_job_local(ctx, event->any.jobId);
	char 		*res[3] = {NULL, NULL, NULL};
	

	/* check auth */
	if (!ctx->isProxy && !ctx->peerName)
		return edg_wll_SetError(ctx,EPERM,"LB server can't store using unauthenticated connection");
	if (ctx->isProxy && (!event->any.user || !strcmp(event->any.user,EDG_WLL_LOG_USER_DEFAULT)) )
		return edg_wll_SetError(ctx,EPERM,"LB proxy can't store using unauthenticated connection");


	trio_asprintf(&q,"select proxy,server,grey from jobs where jobid='%|Ss' for update", unique);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, q);

	nar = edg_wll_ExecSQL(ctx,q,&stmt);
	free(q); q = NULL;

/* XXX: greyjobs semantics is overloaded here:
 * we trust the registration event content to specify job owner unconditionally.
 * It's fine for the time being but should be solved with new authz.
 * */

	if (nar < 0) goto err;
	else if (nar == 0) {
		/* Job not stored yet */
		if (event->any.type == EDG_WLL_EVENT_REGJOB) {

		/* XXX: directness is checked by any.user == peerName. Not perfect but better than event flags. */

			if (!ctx->isProxy && !edg_wll_gss_equal_subj(event->any.user, ctx->peerName) && !ctx->greyjobs) {
				if(ctx->async_registrations) {
					/* check authorization to register job as different user */
					if(check_authz_policy_ctx(ctx, REGISTER_FOREIGN)) {
						edg_wll_SetError(ctx, EPERM, "not allowed to register jobs as different user");
						goto err;
					}
				} else {
					edg_wll_SetError(ctx,EPERM,"job registration must go directly");
					goto err;
				}
			}
			if ((event->any.priority & EDG_WLL_LOGLFLAG_EXCL) && ctx->exclusive_zombies) {
				trio_asprintf(&q,"select jobid from zombie_jobs "
					"where zombie_jobs.jobid='%|Ss'",unique);

				if (edg_wll_ExecSQL(ctx,q,&stmt_zomb) > 0) {
					edg_wll_SetError(ctx,EEXIST,"Zombie job with given jobid already exists. Rejecting event.");
					goto err;
				}
				if (stmt_zomb) { glite_lbu_FreeStmt(&stmt_zomb); stmt_zomb = NULL; }
				free (q); q = NULL;
			}

			/* else OK */
		}
		else {
			if (ctx->greyjobs) grey = 1;
			else if(ctx->async_registrations) { 
				/* TODO: possibly add another permission */
				if(check_authz_policy_ctx(ctx, REGISTER_FOREIGN)) {
					edg_wll_SetError(ctx, EPERM, "not allowed to log events to unregistered job");
					goto err;
				}
				async = 1;
			}
			else {
				edg_wll_SetError(ctx, ENOENT, "job not registered");
				goto err;
			}
		}

		/* first event is not registration:
		 *   - when grey jobs are allowed, job is owned by "GREY JOB"
		 *   - when asynchronous registrations are allowed, job is owned by the logging component
		 * when none of the above applies, we should not get here at all.
		 */
		can_peername = grey ?  strdup("GREY JOB") : ( async ? edg_wll_gss_normalize_subj(ctx->peerName, 0) : edg_wll_gss_normalize_subj(event->any.user, 0) );
		userid = strdup(strmd5(can_peername, NULL));
		if (store_user(ctx,userid,can_peername)) goto err;
		/* store job as unregistered (grey) when either grey jobs or async registrations are allowed */
		if (store_job(ctx,(glite_jobid_const_t) event->any.jobId,
			userid, ctx->isProxy, local_job, grey | async, 0 )) goto err;
		if (event->any.type == EDG_WLL_EVENT_REGJOB) {
			if(event->regJob.wms_dn) {
				char *names = strdup(event->regJob.wms_dn);
				char *p, *name;
				int ret;
				
				name = names;
				while ((p = strchr(name, '\n'))) {
					*p = '\0';
					ret = edg_wll_UpdateACL(ctx, event->any.jobId, name,
								EDG_WLL_CHANGEACL_DN,
								EDG_WLL_CHANGEACL_READ,
								EDG_WLL_CHANGEACL_ALLOW,
								EDG_WLL_CHANGEACL_ADD);
					if (ret) { 
						free(names);
						goto err;
					}
					name = p+1;
				}
				free(names);
			} else if(async) {
				int ret;
				
				/* allow logging component to modify job */
				ret = edg_wll_UpdateACL(ctx, event->any.jobId, 
							strdup(edg_wll_gss_normalize_subj(ctx->peerName, 0)),
							EDG_WLL_CHANGEACL_DN,
							EDG_WLL_CHANGEACL_READ,
							EDG_WLL_CHANGEACL_ALLOW,
							EDG_WLL_CHANGEACL_ADD);
				if (ret) { 
					goto err;
				}
			}
		}

		*register_to_JP = (local_job) ? REG_JOB_TO_JP : 0;
	}
	else {
		/* Job already stored */

		if (edg_wll_FetchRow(ctx,stmt,sizeof(res)/sizeof(res[0]),NULL,res) < 0) goto err;
		if (stmt) { glite_lbu_FreeStmt(&stmt); stmt = NULL; }


		if ((ctx->greyjobs || ctx->async_registrations) && !strcmp(res[2],"1") && 
			(event->any.type == EDG_WLL_EVENT_REGJOB)) 
		{
			/* late registration event for already existing job */

			/* check permission to change owner if not doing greyjobs (ie. in async registration) */
			if(!ctx->greyjobs) {
				
			}

			/* set proper job owner */
			can_peername = edg_wll_gss_normalize_subj(event->any.user, 0);
			userid = strdup(strmd5(can_peername, NULL));
			if (store_user(ctx,userid,can_peername)) goto err;
			if (store_job(ctx,(glite_jobid_const_t) event->any.jobId,
				userid, (ctx->isProxy || !strcmp(res[0],"1")), 
				!strcmp(res[1],"1") || (local_job ? ctx->serverRunning : 0), 0, 1)) goto err;
		}
		else {
			/* check possible server vs. proxy registration ownership clash */
			if (event->any.type == EDG_WLL_EVENT_REGJOB) 
			{
				trio_asprintf(&q,"select u.cert_subj from jobs j, users u "
							"where j.jobid='%|Ss' and u.userid=j.userid",unique);
				glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, q);

				if (edg_wll_ExecSQL(ctx,q,&stmt) < 0
					|| edg_wll_FetchRow(ctx,stmt,1,NULL,&owner) < 0
				) goto err;

				if (!edg_wll_gss_equal_subj(event->any.user, owner)) {
					edg_wll_SetError(ctx,EPERM,"Job already registered with different owner. Rejecting event.");
					goto err;
				}

				if (event->any.priority & EDG_WLL_LOGLFLAG_EXCL) {
					edg_wll_SetError(ctx,EEXIST,"Job with given jobid already exists. Rejecting event.");
					goto err;
				}
			}

			/* update job membership if needed */
			// if (!strcmp(res[0],"1") && !strcmp(res[1],"1") ) /*nothing to do */;
			if ( (!strcmp(res[0],"0") && ctx->isProxy) || (!strcmp(res[1],"0") && !ctx->isProxy) ) {
				trio_asprintf(&q,"update jobs set server='1', proxy='1' where jobid='%|Ss'",
					unique);
				glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, q);
				if (edg_wll_ExecSQL(ctx,q,NULL) < 0) goto err;
				free(q); q = NULL;
			}
		}
	}

err:
	free(res[0]); free(res[1]); free(res[2]);
	if (stmt) glite_lbu_FreeStmt(&stmt);
	free(subj);
	free(userid);
	free(unique);
	if (q) free(q);
	if (owner) free(owner);

	return edg_wll_Error(ctx,NULL,NULL);

}



#define SHORT_LEN	255	/* short_fiels.value db column lenght */

static int store_flesh(edg_wll_Context ctx,edg_wll_Event *e,const char *ulm,char *jobid,int no)
{
#ifndef LB_EVENTS_BLOB
	struct {
		char	*key;
		char	*val;
	} f[20];
	unsigned int	i;
#endif

	char	*stmt;
	int	err = 0, local_ulm = 0;

	assert(ulm || e);
	edg_wll_ResetError(ctx);
	if (!ulm) {
		ulm = edg_wll_UnparseEvent(ctx, e);
		local_ulm = 1;
	}
	if (!ulm) return edg_wll_Error(ctx, NULL, NULL);

#ifndef LB_EVENTS_BLOB
	// in both ways if some preparsed flesh fields in events too
	if (!e) {
		if (edg_wll_ParseEvent(ctx, (char *)ulm, &e) != 0) return edg_wll_Error(ctx, NULL, NULL);
	}

	memset(f,0,sizeof(f)); assert(f[0].key == NULL);

	switch (e->type) {
		case EDG_WLL_EVENT_TRANSFER:
			f[0].key = "DESTINATION";
			f[0].val = edg_wll_SourceToString(e->transfer.destination);
			f[1].key = "DEST_HOST";
			f[1].val = e->transfer.dest_host?strdup(e->transfer.dest_host):NULL;
			f[2].key = "DEST_INSTANCE";
			f[2].val = e->transfer.dest_instance?strdup(e->transfer.dest_instance):NULL;
			f[3].key = "JOB";
			f[3].val = e->transfer.job?strdup(e->transfer.job):NULL;
			f[4].key = "RESULT";
			f[4].val = edg_wll_TransferResultToString(e->transfer.result);
			f[5].key = "REASON";
			f[5].val = e->transfer.reason?strdup(e->transfer.reason):NULL;
			f[6].key = "DEST_JOBID";
			f[6].val = e->transfer.dest_jobid?strdup(e->transfer.dest_jobid):NULL;
			assert(7<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_ACCEPTED:
			f[0].key = "FROM";
			f[0].val = edg_wll_SourceToString(e->accepted.from);
			f[1].key = "FROM_HOST";
			f[1].val = e->accepted.from_host?strdup(e->accepted.from_host):NULL;
			f[2].key = "FROM_INSTANCE";
			f[2].val = e->accepted.from_instance?strdup(e->accepted.from_instance):NULL;
			f[3].key = "LOCAL_JOBID";
			f[3].val = e->accepted.local_jobid?strdup(e->accepted.local_jobid):NULL;
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_REFUSED:
			f[0].key = "FROM";
			f[0].val = edg_wll_SourceToString(e->refused.from);
			f[1].key = "FROM_HOST";
			f[1].val = e->refused.from_host?strdup(e->refused.from_host):NULL;
			f[2].key = "FROM_INSTANCE";
			f[2].val = e->refused.from_instance?strdup(e->refused.from_instance):NULL;
			f[3].key = "REASON";
			f[3].val = e->refused.reason?strdup(e->refused.reason):NULL;
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_ENQUEUED:
			f[0].key = "QUEUE";
			f[0].val = e->enQueued.queue?strdup(e->enQueued.queue):NULL;
			f[1].key = "JOB";
			f[1].val = e->enQueued.job?strdup(e->enQueued.job):NULL;
			f[2].key = "RESULT";
			f[2].val = edg_wll_EnQueuedResultToString(e->enQueued.result);
			f[3].key = "REASON";
			f[3].val = e->enQueued.reason?strdup(e->enQueued.reason):NULL;
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_DEQUEUED:
			f[0].key = "QUEUE";
			f[0].val = e->deQueued.queue?strdup(e->deQueued.queue):NULL;
			f[1].key = "LOCAL_JOBID";
			f[1].val = e->deQueued.local_jobid?strdup(e->deQueued.local_jobid):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_HELPERCALL:
			f[0].key = "HELPER_NAME";
			f[0].val = e->helperCall.helper_name?strdup(e->helperCall.helper_name):NULL;
			f[1].key = "HELPER_PARAMS";
			f[1].val = e->helperCall.helper_params?strdup(e->helperCall.helper_params):NULL;
			f[2].key = "SRC_ROLE";
			f[2].val = edg_wll_HelperCallSrc_roleToString(e->helperCall.src_role);
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_HELPERRETURN:
			f[0].key = "HELPER_NAME";
			f[0].val = e->helperReturn.helper_name?strdup(e->helperReturn.helper_name):NULL;
			f[1].key = "RETVAL";
			f[1].val = e->helperReturn.retval?strdup(e->helperReturn.retval):NULL;
			f[2].key = "SRC_ROLE";
			f[2].val = edg_wll_HelperReturnSrc_roleToString(e->helperReturn.src_role);
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_RUNNING:
			f[0].key = "NODE";
			f[0].val = e->running.node?strdup(e->running.node):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_RESUBMISSION:
			f[0].key = "RESULT";
			f[0].val = edg_wll_ResubmissionResultToString(e->resubmission.result);
			f[1].key = "REASON";
			f[1].val = e->resubmission.reason?strdup(e->resubmission.reason):NULL;
			f[2].key = "TAG";
			f[2].val = e->resubmission.tag?strdup(e->resubmission.tag):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_DONE:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_DoneStatus_codeToString(e->done.status_code);
			f[1].key = "REASON";
			f[1].val = e->done.reason?strdup(e->done.reason):NULL;
			f[2].key = "EXIT_CODE";
			asprintf(&f[2].val,"%d",e->done.exit_code);
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CANCEL:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_CancelStatus_codeToString(e->cancel.status_code);
			f[1].key = "REASON";
			f[1].val = e->cancel.reason?strdup(e->cancel.reason):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_ABORT:
			f[0].key = "REASON";
			f[0].val = e->abort.reason?strdup(e->abort.reason):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CLEAR:
			f[0].key = "REASON";
			f[0].val = edg_wll_ClearReasonToString(e->clear.reason);
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PURGE:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_MATCH:
			f[0].key = "DEST_ID";
			f[0].val = e->match.dest_id?strdup(e->match.dest_id):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PENDING:
			f[0].key = "REASON";
			f[0].val = e->pending.reason?strdup(e->pending.reason):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_REGJOB:
			f[0].key = "JDL";
			f[0].val = e->regJob.jdl?strdup(e->regJob.jdl):NULL;
			f[1].key = "NS";
			f[1].val = e->regJob.ns?strdup(e->regJob.ns):NULL;
			f[2].key = "PARENT";
			f[2].val = edg_wlc_JobIdUnparse(e->regJob.parent);
			f[3].key = "JOBTYPE";
			f[3].val = edg_wll_RegJobJobtypeToString(e->regJob.jobtype);
			f[4].key = "NSUBJOBS";
			asprintf(&f[4].val,"%d",e->regJob.nsubjobs);
			f[5].key = "SEED";
			f[5].val = e->regJob.seed?strdup(e->regJob.seed):NULL;
			f[6].key = "WMS_DN";
			f[6].val = e->regJob.wms_dn?strdup(e->regJob.wms_dn):NULL;
			assert(7<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CHKPT:
			f[0].key = "TAG";
			f[0].val = e->chkpt.tag?strdup(e->chkpt.tag):NULL;
			f[1].key = "CLASSAD";
			f[1].val = e->chkpt.classad?strdup(e->chkpt.classad):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_LISTENER:
			f[0].key = "SVC_NAME";
			f[0].val = e->listener.svc_name?strdup(e->listener.svc_name):NULL;
			f[1].key = "SVC_HOST";
			f[1].val = e->listener.svc_host?strdup(e->listener.svc_host):NULL;
			f[2].key = "SVC_PORT";
			asprintf(&f[2].val,"%d",(int) e->listener.svc_port);
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CURDESCR:
			f[0].key = "DESCR";
			f[0].val = e->curDescr.descr?strdup(e->curDescr.descr):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_USERTAG:
			f[0].key = "NAME";
			f[0].val = e->userTag.name?strdup(e->userTag.name):NULL;
			f[1].key = "VALUE";
			f[1].val = e->userTag.value?strdup(e->userTag.value):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CHANGEACL:
			f[0].key = "USER_ID";
			f[0].val = e->changeACL.user_id?strdup(e->changeACL.user_id):NULL;
			f[1].key = "USER_ID_TYPE";
			f[1].val = edg_wll_ChangeACLUser_id_typeToString(e->changeACL.user_id_type);
			f[2].key = "PERMISSION";
			f[2].val = edg_wll_ChangeACLPermissionToString(e->changeACL.permission);
			f[3].key = "PERMISSION_TYPE";
			f[3].val = edg_wll_ChangeACLPermission_typeToString(e->changeACL.permission_type);
			f[4].key = "OPERATION";
			f[4].val = edg_wll_ChangeACLOperationToString(e->changeACL.operation);
			assert(5<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_NOTIFICATION:
			f[0].key = "NOTIFID";
			f[0].val = edg_wll_NotifIdUnparse(e->notification.notifId);
			f[1].key = "OWNER";
			f[1].val = e->notification.owner?strdup(e->notification.owner):NULL;
			f[2].key = "DEST_HOST";
			f[2].val = e->notification.dest_host?strdup(e->notification.dest_host):NULL;
			f[3].key = "DEST_PORT";
			asprintf(&f[3].val,"%d",(int) e->notification.dest_port);
			f[4].key = "EXPIRES";
			asprintf(&f[4].val,"%d",e->notification.expires);
			f[5].key = "JOBSTAT";
			f[5].val = e->notification.jobstat?strdup(e->notification.jobstat):NULL;
			f[6].key = "DEST_URL";
			f[6].val = e->notification.dest_url?strdup(e->notification.dest_url):NULL;
			assert(7<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_RESOURCEUSAGE:
			f[0].key = "RESOURCE";
			f[0].val = e->resourceUsage.resource?strdup(e->resourceUsage.resource):NULL;
			f[1].key = "QUANTITY";
			asprintf(&f[1].val,"%f",e->resourceUsage.quantity);
			f[2].key = "UNIT";
			f[2].val = e->resourceUsage.unit?strdup(e->resourceUsage.unit):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_REALLYRUNNING:
			f[0].key = "WN_SEQ";
			f[0].val = e->reallyRunning.wn_seq?strdup(e->reallyRunning.wn_seq):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_SUSPEND:
			f[0].key = "REASON";
			f[0].val = e->suspend.reason?strdup(e->suspend.reason):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_RESUME:
			f[0].key = "REASON";
			f[0].val = e->resume.reason?strdup(e->resume.reason):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_COLLECTIONSTATE:
			f[0].key = "STATE";
			f[0].val = e->collectionState.state?strdup(e->collectionState.state):NULL;
			f[1].key = "DONE_CODE";
			asprintf(&f[1].val,"%d",e->collectionState.done_code);
			f[2].key = "HISTOGRAM";
			f[2].val = e->collectionState.histogram?strdup(e->collectionState.histogram):NULL;
			f[3].key = "CHILD";
			f[3].val = edg_wlc_JobIdUnparse(e->collectionState.child);
			f[4].key = "CHILD_EVENT";
			f[4].val = e->collectionState.child_event?strdup(e->collectionState.child_event):NULL;
			assert(5<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_GRANTPAYLOADOWNERSHIP:
			f[0].key = "PAYLOAD_OWNER";
			f[0].val = e->grantPayloadOwnership.payload_owner?strdup(e->grantPayloadOwnership.payload_owner):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_TAKEPAYLOADOWNERSHIP:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONNECTJOB:
			f[0].key = "DEST_JOBID";
			f[0].val = e->connectJob.dest_jobid?strdup(e->connectJob.dest_jobid):NULL;
			f[1].key = "JOBTYPE";
			f[1].val = edg_wll_ConnectJobJobtypeToString(e->connectJob.jobtype);
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_DISCONNECTJOB:
			f[0].key = "DEST_JOBID";
			f[0].val = e->disconnectJob.dest_jobid?strdup(e->disconnectJob.dest_jobid):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSINTERNALSTATECHANGE:
			f[0].key = "OLDSTATE";
			asprintf(&f[0].val,"%d",e->PBSInternalStateChange.oldstate);
			f[1].key = "OLDSUBSTATE";
			asprintf(&f[1].val,"%d",e->PBSInternalStateChange.oldsubstate);
			f[2].key = "NEWSTATE";
			asprintf(&f[2].val,"%d",e->PBSInternalStateChange.newstate);
			f[3].key = "NEWSUBSTATE";
			asprintf(&f[3].val,"%d",e->PBSInternalStateChange.newsubstate);
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSTRANSFER:
			f[0].key = "DESTINATION";
			f[0].val = edg_wll_SourceToString(e->PBSTransfer.destination);
			f[1].key = "DEST_INSTANCE";
			f[1].val = e->PBSTransfer.dest_instance?strdup(e->PBSTransfer.dest_instance):NULL;
			f[2].key = "RESULT";
			f[2].val = edg_wll_PBSTransferResultToString(e->PBSTransfer.result);
			f[3].key = "REASON";
			f[3].val = e->PBSTransfer.reason?strdup(e->PBSTransfer.reason):NULL;
			f[4].key = "DEST_JOBID";
			f[4].val = e->PBSTransfer.dest_jobid?strdup(e->PBSTransfer.dest_jobid):NULL;
			assert(5<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSACCEPTED:
			f[0].key = "FROM";
			f[0].val = edg_wll_SourceToString(e->PBSAccepted.from);
			f[1].key = "FROM_INSTANCE";
			f[1].val = e->PBSAccepted.from_instance?strdup(e->PBSAccepted.from_instance):NULL;
			f[2].key = "LOCAL_JOBID";
			f[2].val = e->PBSAccepted.local_jobid?strdup(e->PBSAccepted.local_jobid):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSREFUSED:
			f[0].key = "FROM";
			f[0].val = edg_wll_SourceToString(e->PBSRefused.from);
			f[1].key = "FROM_INSTANCE";
			f[1].val = e->PBSRefused.from_instance?strdup(e->PBSRefused.from_instance):NULL;
			f[2].key = "REASON";
			f[2].val = e->PBSRefused.reason?strdup(e->PBSRefused.reason):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSQUEUED:
			f[0].key = "QUEUE";
			f[0].val = e->PBSQueued.queue?strdup(e->PBSQueued.queue):NULL;
			f[1].key = "NAME";
			f[1].val = e->PBSQueued.name?strdup(e->PBSQueued.name):NULL;
			f[2].key = "OWNER";
			f[2].val = e->PBSQueued.owner?strdup(e->PBSQueued.owner):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSDEQUEUED:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSMATCH:
			f[0].key = "DEST_HOST";
			f[0].val = e->PBSMatch.dest_host?strdup(e->PBSMatch.dest_host):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSPENDING:
			f[0].key = "REASON";
			f[0].val = e->PBSPending.reason?strdup(e->PBSPending.reason):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSWAITING:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSRUN:
			f[0].key = "SCHEDULER";
			f[0].val = e->PBSRun.scheduler?strdup(e->PBSRun.scheduler):NULL;
			f[1].key = "DEST_HOST";
			f[1].val = e->PBSRun.dest_host?strdup(e->PBSRun.dest_host):NULL;
			f[2].key = "PID";
			asprintf(&f[2].val,"%d",e->PBSRun.pid);
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSRERUN:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSABORT:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSDONE:
			f[0].key = "EXIT_STATUS";
			asprintf(&f[0].val,"%d",e->PBSDone.exit_status);
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSRESOURCEUSAGE:
			f[0].key = "USAGE";
			f[0].val = edg_wll_PBSResourceUsageUsageToString(e->PBSResourceUsage.usage);
			f[1].key = "RESOURCES";
			f[1].val = edg_wll_TagListToString(e->PBSResourceUsage.resources);
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_PBSERROR:
			f[0].key = "ERROR_DESC";
			f[0].val = e->PBSError.error_desc?strdup(e->PBSError.error_desc):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORMATCH:
			f[0].key = "OWNER";
			f[0].val = e->CondorMatch.owner?strdup(e->CondorMatch.owner):NULL;
			f[1].key = "DEST_HOST";
			f[1].val = e->CondorMatch.dest_host?strdup(e->CondorMatch.dest_host):NULL;
			f[2].key = "PREEMPTING";
			f[2].val = e->CondorMatch.preempting?strdup(e->CondorMatch.preempting):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORREJECT:
			f[0].key = "OWNER";
			f[0].val = e->CondorReject.owner?strdup(e->CondorReject.owner):NULL;
			f[1].key = "STATUS_CODE";
			f[1].val = edg_wll_CondorRejectStatus_codeToString(e->CondorReject.status_code);
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORSHADOWSTARTED:
			f[0].key = "SHADOW_HOST";
			f[0].val = e->CondorShadowStarted.shadow_host?strdup(e->CondorShadowStarted.shadow_host):NULL;
			f[1].key = "SHADOW_PORT";
			asprintf(&f[1].val,"%d",(int) e->CondorShadowStarted.shadow_port);
			f[2].key = "SHADOW_PID";
			asprintf(&f[2].val,"%d",e->CondorShadowStarted.shadow_pid);
			f[3].key = "SHADOW_STATUS";
			f[3].val = e->CondorShadowStarted.shadow_status?strdup(e->CondorShadowStarted.shadow_status):NULL;
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORSHADOWEXITED:
			f[0].key = "SHADOW_PID";
			asprintf(&f[0].val,"%d",e->CondorShadowExited.shadow_pid);
			f[1].key = "SHADOW_EXIT_STATUS";
			asprintf(&f[1].val,"%d",e->CondorShadowExited.shadow_exit_status);
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORSTARTERSTARTED:
			f[0].key = "STARTER_PID";
			asprintf(&f[0].val,"%d",e->CondorStarterStarted.starter_pid);
			f[1].key = "UNIVERSE";
			f[1].val = e->CondorStarterStarted.universe?strdup(e->CondorStarterStarted.universe):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORSTARTEREXITED:
			f[0].key = "STARTER_PID";
			asprintf(&f[0].val,"%d",e->CondorStarterExited.starter_pid);
			f[1].key = "STARTER_EXIT_STATUS";
			asprintf(&f[1].val,"%d",e->CondorStarterExited.starter_exit_status);
			f[2].key = "JOB_PID";
			asprintf(&f[2].val,"%d",e->CondorStarterExited.job_pid);
			f[3].key = "JOB_EXIT_STATUS";
			asprintf(&f[3].val,"%d",e->CondorStarterExited.job_exit_status);
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORRESOURCEUSAGE:
			f[0].key = "USAGE";
			f[0].val = edg_wll_CondorResourceUsageUsageToString(e->CondorResourceUsage.usage);
			f[1].key = "NAME";
			f[1].val = e->CondorResourceUsage.name?strdup(e->CondorResourceUsage.name):NULL;
			f[2].key = "QUANTITY";
			asprintf(&f[2].val,"%f",e->CondorResourceUsage.quantity);
			f[3].key = "UNIT";
			f[3].val = e->CondorResourceUsage.unit?strdup(e->CondorResourceUsage.unit):NULL;
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CONDORERROR:
			f[0].key = "ERROR_DESC";
			f[0].val = e->CondorError.error_desc?strdup(e->CondorError.error_desc):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMSTART:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMPURGE:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMACCEPTED:
			f[0].key = "LOCAL_JOBID";
			f[0].val = e->CREAMAccepted.local_jobid?strdup(e->CREAMAccepted.local_jobid):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMSTORE:
			f[0].key = "COMMAND";
			f[0].val = edg_wll_CREAMStoreCommandToString(e->CREAMStore.command);
			f[1].key = "CMDID";
			f[1].val = e->CREAMStore.cmdid?strdup(e->CREAMStore.cmdid):NULL;
			f[2].key = "RESULT";
			f[2].val = edg_wll_CREAMStoreResultToString(e->CREAMStore.result);
			f[3].key = "REASON";
			f[3].val = e->CREAMStore.reason?strdup(e->CREAMStore.reason):NULL;
			assert(4<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMCALL:
			f[0].key = "COMMAND";
			f[0].val = edg_wll_CREAMCallCommandToString(e->CREAMCall.command);
			f[1].key = "CALLEE";
			f[1].val = edg_wll_SourceToString(e->CREAMCall.callee);
			f[2].key = "CMDID";
			f[2].val = e->CREAMCall.cmdid?strdup(e->CREAMCall.cmdid):NULL;
			f[3].key = "DESTID";
			f[3].val = e->CREAMCall.destid?strdup(e->CREAMCall.destid):NULL;
			f[4].key = "RESULT";
			f[4].val = edg_wll_CREAMCallResultToString(e->CREAMCall.result);
			f[5].key = "REASON";
			f[5].val = e->CREAMCall.reason?strdup(e->CREAMCall.reason):NULL;
			assert(6<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMRUNNING:
			f[0].key = "NODE";
			f[0].val = e->CREAMRunning.node?strdup(e->CREAMRunning.node):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMREALLYRUNNING:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMDONE:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_CREAMDoneStatus_codeToString(e->CREAMDone.status_code);
			f[1].key = "REASON";
			f[1].val = e->CREAMDone.reason?strdup(e->CREAMDone.reason):NULL;
			f[2].key = "EXIT_CODE";
			f[2].val = e->CREAMDone.exit_code?strdup(e->CREAMDone.exit_code):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMCANCEL:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_CREAMCancelStatus_codeToString(e->CREAMCancel.status_code);
			f[1].key = "REASON";
			f[1].val = e->CREAMCancel.reason?strdup(e->CREAMCancel.reason):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMABORT:
			f[0].key = "REASON";
			f[0].val = e->CREAMAbort.reason?strdup(e->CREAMAbort.reason):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMSTATUS:
			f[0].key = "OLD_STATE";
			f[0].val = e->CREAMStatus.old_state?strdup(e->CREAMStatus.old_state):NULL;
			f[1].key = "NEW_STATE";
			f[1].val = e->CREAMStatus.new_state?strdup(e->CREAMStatus.new_state):NULL;
			f[2].key = "ORIG_TIMESTAMP";
			edg_wll_ULMTimevalToDate((e->CREAMStatus.orig_timestamp).tv_sec,(e->CREAMStatus.orig_timestamp).tv_usec,f[2].val);
			f[3].key = "RESULT";
			f[3].val = edg_wll_CREAMStatusResultToString(e->CREAMStatus.result);
			f[4].key = "DESCR";
			f[4].val = e->CREAMStatus.descr?strdup(e->CREAMStatus.descr):NULL;
			f[5].key = "EXIT_CODE";
			f[5].val = e->CREAMStatus.exit_code?strdup(e->CREAMStatus.exit_code):NULL;
			f[6].key = "FAILURE_REASON";
			f[6].val = e->CREAMStatus.failure_reason?strdup(e->CREAMStatus.failure_reason):NULL;
			f[7].key = "WORKER_NODE";
			f[7].val = e->CREAMStatus.worker_node?strdup(e->CREAMStatus.worker_node):NULL;
			f[8].key = "LRMS_JOBID";
			f[8].val = e->CREAMStatus.LRMS_jobid?strdup(e->CREAMStatus.LRMS_jobid):NULL;
			assert(9<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMSUSPEND:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_CREAMSuspendStatus_codeToString(e->CREAMSuspend.status_code);
			f[1].key = "REASON";
			f[1].val = e->CREAMSuspend.reason?strdup(e->CREAMSuspend.reason):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_CREAMRESUME:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_CREAMResumeStatus_codeToString(e->CREAMResume.status_code);
			f[1].key = "REASON";
			f[1].val = e->CREAMResume.reason?strdup(e->CREAMResume.reason):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_FILETRANSFERREGISTER:
			f[0].key = "SRC";
			f[0].val = e->fileTransferRegister.src?strdup(e->fileTransferRegister.src):NULL;
			f[1].key = "DEST";
			f[1].val = e->fileTransferRegister.dest?strdup(e->fileTransferRegister.dest):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_FILETRANSFER:
			f[0].key = "RESULT";
			f[0].val = edg_wll_FileTransferResultToString(e->fileTransfer.result);
			f[1].key = "REASON";
			f[1].val = e->fileTransfer.reason?strdup(e->fileTransfer.reason):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_SANDBOX:
			f[0].key = "SANDBOX_TYPE";
			f[0].val = edg_wll_SandboxSandbox_typeToString(e->sandbox.sandbox_type);
			f[1].key = "TRANSFER_JOB";
			f[1].val = e->sandbox.transfer_job?strdup(e->sandbox.transfer_job):NULL;
			f[2].key = "COMPUTE_JOB";
			f[2].val = e->sandbox.compute_job?strdup(e->sandbox.compute_job):NULL;
			assert(3<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMCREATE:
			f[0].key = "REQUIRE";
			f[0].val = e->vMCreate.require?strdup(e->vMCreate.require):NULL;
			f[1].key = "ID";
			f[1].val = e->vMCreate.id?strdup(e->vMCreate.id):NULL;
			f[2].key = "NAME";
			f[2].val = e->vMCreate.name?strdup(e->vMCreate.name):NULL;
			f[3].key = "OWNER";
			f[3].val = e->vMCreate.owner?strdup(e->vMCreate.owner):NULL;
			f[4].key = "HOSTNAME";
			f[4].val = e->vMCreate.hostname?strdup(e->vMCreate.hostname):NULL;
			f[5].key = "MANAGER";
			f[5].val = e->vMCreate.manager?strdup(e->vMCreate.manager):NULL;
			f[6].key = "IMAGE";
			f[6].val = e->vMCreate.image?strdup(e->vMCreate.image):NULL;
			assert(7<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMHOST:
			f[0].key = "HOSTNAME";
			f[0].val = e->vMHost.hostname?strdup(e->vMHost.hostname):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMIMAGE:
			f[0].key = "IMAGE";
			f[0].val = e->vMImage.image?strdup(e->vMImage.image):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMRUNNING:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMSHUTDOWN:
			f[0].key = "USAGE";
			f[0].val = e->vMShutdown.usage?strdup(e->vMShutdown.usage):NULL;
			assert(1<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMSTOP:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMRESUME:
			assert(0<sizeof f/sizeof f[0]);
			break;
		case EDG_WLL_EVENT_VMDONE:
			f[0].key = "STATUS_CODE";
			f[0].val = edg_wll_VMDoneStatus_codeToString(e->vMDone.status_code);
			f[1].key = "USAGE";
			f[1].val = e->vMDone.usage?strdup(e->vMDone.usage):NULL;
			assert(2<sizeof f/sizeof f[0]);
			break;
		default:
			break;
	}

	for (i=0; i<sizeof(f)/sizeof(f[0]) && !err; i++) if (f[i].key && f[i].val) {
		trio_asprintf(&stmt,"insert into %s(jobid,event,name,value) "
			"values ('%|Ss',%d,'%|Ss','%|Ss')",
			strlen(f[i].val) <= SHORT_LEN ? "short_fields" : "long_fields",
			jobid,no,f[i].key,f[i].val);
		glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

		if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) err = edg_wll_Error(ctx,NULL,NULL);
		free(stmt);
	}
	for (i=0; i<sizeof(f)/sizeof(f[0]); i++) free(f[i].val);

/* XXX: hardcoded, no other suitable place to store it */
	if (!err) {
		trio_asprintf(&stmt,"insert into short_fields(jobid,event,name,value) "
			"values ('%|Ss',%d,'SRC_INSTANCE','%|Ss')",
			jobid,no,e->any.src_instance);
		glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

		if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) err = edg_wll_Error(ctx,NULL,NULL);
		free(stmt);
	}
#else
	trio_asprintf(&stmt, "insert into events_flesh (jobid, event, ulm) values ('%|Ss', %d, '%|Ss')", jobid, no, ulm);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,NULL) < 0) err = edg_wll_Error(ctx,NULL,NULL);
	free(stmt);
#endif

	if (local_ulm) free(ulm);

	return err;
}

static int check_dup(edg_wll_Context ctx,edg_wll_Event *e)
{
	int		i,dup_detected = 0;
	int		err;
	char		*es,*es2;
	edg_wll_QueryRec	jc[2],ec[2];
	edg_wll_QueryRec        **jca, **eca;
	edg_wll_Event	*e2;

	edg_wll_ResetError(ctx);

	jc[0].attr = EDG_WLL_QUERY_ATTR_JOBID;
	jc[0].value.j = e->any.jobId;
	jc[0].op = EDG_WLL_QUERY_OP_EQUAL;
	jc[1].attr = EDG_WLL_QUERY_ATTR_UNDEF;

	ec[0].attr = EDG_WLL_QUERY_ATTR_TIME;
	memcpy(&ec[0].value.t,&e->any.timestamp,sizeof(struct timeval));
	ec[0].op = EDG_WLL_QUERY_OP_EQUAL;
	ec[1].attr = EDG_WLL_QUERY_ATTR_UNDEF;

        jca = (edg_wll_QueryRec **) malloc (2 * sizeof(edg_wll_QueryRec **));
        eca = (edg_wll_QueryRec **) malloc (2 * sizeof(edg_wll_QueryRec **));
        jca[0] = jc;
        jca[1] = NULL;
        eca[0] = ec;
        eca[1] = NULL;

	err = edg_wll_QueryEventsServer(ctx,1,(const edg_wll_QueryRec **)jca, 
					(const edg_wll_QueryRec **)eca,&e2);
	switch (err) {
		case 0: /* continue normally */
			break;
		case ENOENT:
		case EIDRM:
			free(jca);
			free(eca);
			return edg_wll_ResetError(ctx);
			break;
		default:
			free(jca);
			free(eca);
			return edg_wll_Error(ctx,NULL,NULL);
			break;
	}

	es = edg_wll_UnparseEvent(ctx,e);
	assert(es);

	for (i=0;e2[i].type && !dup_detected;i++) {
		/* Ignore priority */
		e2[i].any.priority = e->any.priority;
		es2 = edg_wll_UnparseEvent(ctx,e2+i);
		assert(es2);
		if (!strcmp(es,es2)) {
			dup_detected = 1;
			edg_wll_SetError(ctx,EEXIST,"duplicate event");
		}
		free(es2);
	}

	free(jca);
	free(eca);
	free(es);
	for (i=0; e2[i].type; i++) edg_wll_FreeEvent(e2+i);
	free(e2);
	return edg_wll_Error(ctx,NULL,NULL);
}

static int
check_acl(edg_wll_Context ctx,edg_wll_Event *e, enum edg_wll_ChangeACLPermission p)
{
	int ret;
	edg_wll_Acl acl = NULL;
	struct _edg_wll_GssPrincipal_data princ;

	memset(&princ, 0, sizeof(princ));

	ret = edg_wll_GetACL(ctx, e->any.jobId, &acl);
	if (ret)
		return EPERM;

	/* XXX don't have VOMS attributes here */
	princ.name = e->any.user;
	ret = edg_wll_CheckACL_princ(ctx, acl, p, &princ);
	edg_wll_FreeAcl(acl);
	return ret;
}

static int check_auth(edg_wll_Context ctx,edg_wll_Event *e)
{
	char	*jobid = NULL;
	char	*q = NULL, *owner = NULL;
	glite_lbu_Statement	stmt = NULL;

	edg_wll_ResetError(ctx);

	if (!ctx->isProxy && !ctx->peerName) 
		return edg_wll_SetError(ctx,EPERM,"can't store using unauthenticated connection");

	jobid = edg_wlc_JobIdGetUnique(e->any.jobId);
	trio_asprintf(&q,"select u.cert_subj from jobs j, users u "
				"where j.jobid='%|Ss' and u.userid=j.userid",jobid);
	free(jobid);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, q);

	if (edg_wll_ExecSQL(ctx,q,&stmt) < 0
		|| edg_wll_FetchRow(ctx,stmt,1,NULL,&owner) < 0
	) goto clean;

	if (!owner) {
		edg_wll_SetError(ctx, ENOENT, "job not registered");
		goto clean;
	}

	switch (e->any.source) {
		case EDG_WLL_SOURCE_USER_INTERFACE:
		case EDG_WLL_SOURCE_LRMS:
		case EDG_WLL_SOURCE_APPLICATION:
			if (edg_wll_gss_equal_subj(owner,e->any.user))
				break;
			if (e->any.type == EDG_WLL_EVENT_USERTAG &&
			    check_acl(ctx, e, EDG_WLL_CHANGEACL_TAG) == 0)
				break;
			if (e->any.type == EDG_WLL_EVENT_TAKEPAYLOADOWNERSHIP &&
			    strcmp(e->any.user,EDG_WLL_LOG_USER_DEFAULT) != 0)
				break;
		
			edg_wll_SetError(ctx,EPERM,"check_auth()");
			break;
		default:
			/* XXX: just don't allow anonymous */
			if (!strcmp(e->any.user,EDG_WLL_LOG_USER_DEFAULT))
				edg_wll_SetError(ctx,EPERM,"check_auth()");
			break;
	}


clean:
	if (stmt) glite_lbu_FreeStmt(&stmt);
	free(q);
	free(owner);
	return edg_wll_Error(ctx,NULL,NULL);
}


edg_wll_ErrorCode intJobStat_embryonic(
	edg_wll_Context ctx,
	glite_jobid_const_t jobid,
	const edg_wll_RegJobEvent *e,
	intJobStat *stat)
{
	if (edg_wlc_JobIdDup(jobid, &intJobStat_to_JobStat(stat)->jobId) ||
		edg_wlc_JobIdDup(e->jobId, &intJobStat_to_JobStat(stat)->parent_job)) goto err;
	intJobStat_to_JobStat(stat)->state = EDG_WLL_JOB_SUBMITTED;
	intJobStat_to_JobStat(stat)->owner = strdup(e->user);
	if (e->jobtype == EDG_WLL_REGJOB_FILE_TRANSFER_COLLECTION)
		intJobStat_to_JobStat(stat)->jobtype = EDG_WLL_STAT_FILE_TRANSFER;
	else /* be defensive here */
		intJobStat_to_JobStat(stat)->jobtype = EDG_WLL_STAT_SIMPLE;
	intJobStat_to_JobStat(stat)->stateEnterTimes[1 + EDG_WLL_JOB_SUBMITTED] = (int)e->timestamp.tv_sec;
	intJobStat_to_JobStat(stat)->lastUpdateTime = e->timestamp;
	intJobStat_to_JobStat(stat)->stateEnterTime = e->timestamp;

err:
	return edg_wll_Error(ctx,NULL,NULL);
}

/*
 * Returns encoded SQL table states record for embryonic DAG subjob
 */

static edg_wll_ErrorCode states_values_embryonic(
	edg_wll_Context ctx,
	glite_jobid_const_t jobid,
	const edg_wll_RegJobEvent *e,
	char **icnames,
	char **values)
{
	char *jobid_md5, *stat_enc, *parent_md5;
	char *stmt = NULL;
	char *icvalues;
	intJobStat stat_rec;
	intJobStat *stat = &stat_rec;

	init_intJobStat(stat);

	if (intJobStat_embryonic(ctx, jobid, e, stat)) goto err;
	
	jobid_md5 = edg_wlc_JobIdGetUnique(jobid);
	parent_md5 = edg_wlc_JobIdGetUnique(e->jobId);
	stat_enc = enc_intJobStat(strdup(""), stat);
	if (jobid_md5 == NULL || parent_md5 == NULL || stat_enc == NULL) goto err;


	if (edg_wll_IColumnsSQLPart(ctx, ctx->job_index_cols, intJobStat_to_JobStat(stat), 1, icnames, &icvalues)) goto err;
	trio_asprintf(&stmt,
		"'%|Ss',%d,%d,'%|Ss','%|Ss','%|Ss'%s",
		jobid_md5, intJobStat_to_JobStat(stat)->state, 1, stat_enc,
		INTSTAT_VERSION, parent_md5, icvalues);
	free(icvalues);

err:
	destroy_intJobStat(stat);
	free(jobid_md5);
	free(stat_enc);
	free(parent_md5);
	*values = stmt;
	return edg_wll_Error(ctx,NULL,NULL);
}


int register_subjobs_embryonic(edg_wll_Context ctx,const edg_wll_RegJobEvent *e)
{
	int			i, j,  err = 0;
	edg_wlc_JobId		*subjobs = NULL;
	char			*jobid = NULL, *jobid_md5 = NULL, *jobid_md5_old = NULL;
	size_t			jobid_len;
	char			*icnames = NULL, *values = NULL, *userid = NULL, *stmt = NULL;
	int			server, proxy, membership = 0;
	glite_lbu_Statement     sh = NULL;


	edg_wll_ResetError(ctx);

	if (e->nsubjobs == 0) return 0;
	if (e->nsubjobs < 0) return edg_wll_SetError(ctx,EINVAL,"negative number of subjobs");
	if ((err = edg_wll_GenerateSubjobIds(ctx,e->jobId,e->nsubjobs,e->seed,&subjobs)))
		return err;

	/* find out icnames and values once, then only change jobids */ 
	if (states_values_embryonic(ctx, subjobs[0], e, &icnames, &values))
		edg_wll_Error(ctx, NULL, NULL);
	jobid_md5_old = edg_wlc_JobIdGetUnique(subjobs[0]);
	jobid_len = strlen(jobid_md5_old);
		

	/* increase the overall request timeout. */
	ctx->p_tmp_timeout.tv_sec += e->nsubjobs/10;
	if (ctx->p_tmp_timeout.tv_sec > 86400) ctx->p_tmp_timeout.tv_sec = 86400;

	if ((membership = edg_wll_jobMembership(ctx, e->jobId)) < 0) goto err;

	proxy = membership & DB_PROXY_JOB;
	server = membership & DB_SERVER_JOB;

	/* get userid of parent job */
	jobid = edg_wlc_JobIdGetUnique(e->jobId);
	trio_asprintf(&stmt,"select userid from jobs where jobid='%|Ss'", jobid);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	if (edg_wll_ExecSQL(ctx,stmt,&sh) < 0 || edg_wll_FetchRow(ctx,sh,1,NULL,&userid) < 0) goto err;

	for (i=0; i<e->nsubjobs; i++) {
		char		*et,*ed,*job_s,*p,*p1;

		/* save jobid-userid relation into jobs table */
	        if ((err = store_job(ctx, subjobs[i], userid, proxy, server, 0, 0)))
			if (edg_wll_Error(ctx,&et,&ed) == EDEADLK) goto err;

		/* interchange variable parts (jobids) in values */
		/* there are only two occurences of subjob jobid */
		jobid_md5 = edg_wlc_JobIdGetUnique(subjobs[i]);
		if (i) {
			p = strstr(values, jobid_md5_old);		
			assert(p);
			memcpy(p, jobid_md5, jobid_len);

			p1 = strstr(p + jobid_len, jobid_md5_old);
			assert(p1);
			memcpy(p1, jobid_md5, jobid_len);
		}
		free(jobid_md5_old);
		jobid_md5_old = jobid_md5;

		if (!err && (err = edg_wll_StoreIntStateEmbryonic(ctx, icnames, values)))
			edg_wll_Error(ctx,&et,&ed);

		if (err) {
			job_s = edg_wlc_JobIdUnparse(subjobs[i]);
			glite_common_log(LOG_CATEGORY_CONTROL, LOG_PRIORITY_ERROR, "%s: %s (%s)", job_s, et, ed);
			free(job_s); free(et); free(ed);
			edg_wll_ResetError(ctx);
		}
		edg_wlc_JobIdFree(subjobs[i]);
	}

err:
	free(jobid_md5_old);	//free the last one
	free(icnames);
	free(values);
	/* free the rest of subjobs if DEADLOCK occurs */
	for (j=i; j<e->nsubjobs; j++) edg_wlc_JobIdFree(subjobs[j]);
	free(subjobs);
	if (sh) glite_lbu_FreeStmt(&sh);
	free(stmt);

	return edg_wll_Error(ctx,NULL,NULL);
}


int edg_wll_delete_event(edg_wll_Context ctx,const char *jobid,int event)
{
	char	*stmt;

/* The order of tables is important to prevent another process calling
 * StoreEvent() to get our event number and mess up the fields together.
 *
 * XXX: best effort: more or less ignore errors
 *
 */

	trio_asprintf(&stmt,
		"delete from short_fields where jobid='%|Ss' and event=%d",
		jobid,event);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	edg_wll_ExecSQL(ctx,stmt,NULL);
	free(stmt);

	trio_asprintf(&stmt,
		"delete from long_fields where jobid='%|Ss' and event=%d",
		jobid,event);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	edg_wll_ExecSQL(ctx,stmt,NULL);
	free(stmt);

	trio_asprintf(&stmt,
		"delete from events_flesh where jobid='%|Ss' and event=%d",
		jobid,event);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	edg_wll_ExecSQL(ctx,stmt,NULL);
	free(stmt);

	trio_asprintf(&stmt,
		"delete from events where jobid='%|Ss' and event=%d",
		jobid,event);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	edg_wll_ExecSQL(ctx,stmt,NULL);
	free (stmt);

	return edg_wll_Error(ctx,NULL,NULL);
}


/* XXX: if event type is user tag, convert the tag name to lowercase!
 * 	  (not sure whether to convert a value too is reasonable
 * 	  or keep it 'case sensitive')
 */
static void lowercase_usertag(edg_wll_Event *ev)
{
	int i;

	if ( ev->any.type == EDG_WLL_EVENT_USERTAG ) {

		for ( i = 0; ev->userTag.name[i] != '\0'; i++ )
		ev->userTag.name[i] = tolower(ev->userTag.name[i]);
	}
}

int edg_wll_jobsconnection_create(edg_wll_Context ctx, glite_jobid_const_t jobid_from, glite_jobid_const_t jobid_to, enum edg_wll_StatJobtype jobtype, edg_wll_JobConnectionType connectiontype){
        char *stmt;

	char *jobid_from_u, *jobid_to_u;
	jobid_from_u = edg_wlc_JobIdGetUnique(jobid_from);
	jobid_to_u = edg_wlc_JobIdGetUnique(jobid_to);

	// clear previous connection of same jobs (e.g. resubmition on same machine)
	trio_asprintf(&stmt,
		"delete from job_connections where jobid_from='%s' and jobid_to='%s'",
		jobid_from_u, jobid_to_u
	);

	edg_wll_ExecSQL(ctx,stmt,NULL);

	free(stmt);

        trio_asprintf(&stmt,
                "insert into job_connections (jobid_from, jobid_to, jobtype, connection) values ('%s', '%s', %i, %i)", 
		jobid_from_u, jobid_to_u, jobtype, connectiontype
        );
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);
	free(jobid_from_u);
	free(jobid_to_u);

        edg_wll_ExecSQL(ctx,stmt,NULL);

        free(stmt);

        return edg_wll_Error(ctx, NULL, NULL);
}

int edg_wll_jobsconnection_modify(edg_wll_Context ctx, glite_jobid_const_t jobid_from, glite_jobid_const_t jobid_to, edg_wll_JobConnectionType connectiontype){
        char *stmt;

        char *jobid_from_u, *jobid_to_u;
        jobid_from_u = edg_wlc_JobIdGetUnique(jobid_from);
        jobid_to_u = edg_wlc_JobIdGetUnique(jobid_to);

        trio_asprintf(&stmt,
                "update job_connections set connection=%i where jobid_from='%s' and jobid_to='%s'",
                connectiontype, jobid_from_u, jobid_to_u
        );
        glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);
        free(jobid_from_u);
        free(jobid_to_u);

        edg_wll_ExecSQL(ctx,stmt,NULL);

        free(stmt);

        return edg_wll_Error(ctx, NULL, NULL);
}

int edg_wll_jobsconnection_modifyall(edg_wll_Context ctx, glite_jobid_const_t jobid, edg_wll_JobConnectionType oldtype, edg_wll_JobConnectionType newtype){
        char *stmt;

        char *jobid_u;
        jobid_u = edg_wlc_JobIdGetUnique(jobid);

        trio_asprintf(&stmt,
                "update job_connections set connection=%i where connection=%i and (jobid_from='%s' or jobid_to='%s')",
                newtype, oldtype, jobid_u, jobid_u
        );
        glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);
        free(jobid_u);

        edg_wll_ExecSQL(ctx,stmt,NULL);

        free(stmt);

        return edg_wll_Error(ctx, NULL, NULL);
}


int edg_wll_jobsconnection_purgeall(edg_wll_Context ctx, glite_jobid_const_t jobid){
	char *stmt;

	char *jobid_u;
	jobid_u = edg_wlc_JobIdGetUnique(jobid);

	trio_asprintf(&stmt,
		"delete from job_connections where jobid_from='%s' or jobid_to='%s'",
		jobid_u, jobid_u
	);
	glite_common_log_msg(LOG_CATEGORY_LB_SERVER_DB, LOG_PRIORITY_DEBUG, stmt);

	free(jobid_u);

	edg_wll_ExecSQL(ctx,stmt,NULL);

	free(stmt);

	return edg_wll_Error(ctx, NULL, NULL);
}

